#include <Wire.h>
#include "Adafruit_MCP23017.h"
#include <LiquidTWI.h>
#include <SoftwareSerial.h>

/*Constants**************************************************************************************************************/
const int iPairSignal = 3;
const int oPairBtn = 4;
const int oLowBatLED = 5;
const int oPairLED = 6;
const int iVBat = 14;
const int oPowerBtnLED = 8;
const int oModeBtnLED = 9;
const int oEscapeBtnLED = 12;

// Key modifier definitions
const uint8_t MODIFIER_NONE          =byte((0));
const uint8_t MODIFIER_CONTROL_LEFT  =byte((1<<0));
const uint8_t MODIFIER_SHIFT_LEFT    =byte((1<<1));
const uint8_t MODIFIER_ALT_LEFT      =byte((1<<2));
const uint8_t MODIFIER_GUI_LEFT      =byte((1<<3));
const uint8_t MODIFIER_CONTROL_RIGHT =byte((1<<4));
const uint8_t MODIFIER_SHIFT_RIGHT   =byte((1<<5));
const uint8_t MODIFIER_ALT_RIGHT     =byte((1<<6));
const uint8_t MODIFIER_GUI_RIGHT     =byte((1<<7));

// Keycode definitions
const uint8_t KEY_NONE               =byte(0x00);
const uint8_t KEY_A                  =byte(0x04);
const uint8_t KEY_B                  =byte(0x05);
const uint8_t KEY_C                  =byte(0x06);
const uint8_t KEY_D                  =byte(0x07);
const uint8_t KEY_E                  =byte(0x08);
const uint8_t KEY_F                  =byte(0x09);
const uint8_t KEY_G                  =byte(0x0A);
const uint8_t KEY_H                  =byte(0x0B);
const uint8_t KEY_I                  =byte(0x0C);
const uint8_t KEY_J                  =byte(0x0D);
const uint8_t KEY_K                  =byte(0x0E);
const uint8_t KEY_L                  =byte(0x0F);
const uint8_t KEY_M                  =byte(0x10);
const uint8_t KEY_N                  =byte(0x11);
const uint8_t KEY_O                  =byte(0x12);
const uint8_t KEY_P                  =byte(0x13);
const uint8_t KEY_Q                  =byte(0x14);
const uint8_t KEY_R                  =byte(0x15);
const uint8_t KEY_S                  =byte(0x16);
const uint8_t KEY_T                  =byte(0x17);
const uint8_t KEY_U                  =byte(0x18);
const uint8_t KEY_V                  =byte(0x19);
const uint8_t KEY_W                  =byte(0x1A);
const uint8_t KEY_X                  =byte(0x1B);
const uint8_t KEY_Y                  =byte(0x1C);
const uint8_t KEY_Z                  =byte(0x1D);
const uint8_t KEY_1                  =byte(0x1E);
const uint8_t KEY_2                  =byte(0x1F);
const uint8_t KEY_3                  =byte(0x20);
const uint8_t KEY_4                  =byte(0x21);
const uint8_t KEY_5                  =byte(0x22);
const uint8_t KEY_6                  =byte(0x23);
const uint8_t KEY_7                  =byte(0x24);
const uint8_t KEY_8                  =byte(0x25);
const uint8_t KEY_9                  =byte(0x26);
const uint8_t KEY_0                  =byte(0x27);
const uint8_t KEY_RETURN             =byte(0x28);
const uint8_t KEY_ESCAPE             =byte(0x29);
const uint8_t KEY_BACKSPACE          =byte(0x2A);
const uint8_t KEY_TAB                =byte(0x2B);
const uint8_t KEY_SPACE              =byte(0x2C);
const uint8_t KEY_MINUS              =byte(0x2D);
const uint8_t KEY_EQUAL              =byte(0x2E);
const uint8_t KEY_BRACKET_LEFT       =byte(0x2F);
const uint8_t KEY_BRACKET_RIGHT      =byte(0x30);
const uint8_t KEY_BACKSLASH          =byte(0x31);
const uint8_t KEY_EUROPE_1           =byte(0x32);
const uint8_t KEY_SEMICOLON          =byte(0x33);
const uint8_t KEY_APOSTROPHE         =byte(0x34);
const uint8_t KEY_GRAVE              =byte(0x35);
const uint8_t KEY_COMMA              =byte(0x36);
const uint8_t KEY_PERIOD             =byte(0x37);
const uint8_t KEY_SLASH              =byte(0x38);
const uint8_t KEY_CAPS_LOCK          =byte(0x39);
const uint8_t KEY_F1                 =byte(0x3A);
const uint8_t KEY_F2                 =byte(0x3B);
const uint8_t KEY_F3                 =byte(0x3C);
const uint8_t KEY_F4                 =byte(0x3D);
const uint8_t KEY_F5                 =byte(0x3E);
const uint8_t KEY_F6                 =byte(0x3F);
const uint8_t KEY_F7                 =byte(0x40);
const uint8_t KEY_F8                 =byte(0x41);
const uint8_t KEY_F9                 =byte(0x42);
const uint8_t KEY_F10                =byte(0x43);
const uint8_t KEY_F11                =byte(0x44);
const uint8_t KEY_F12                =byte(0x45);
const uint8_t KEY_PRINT_SCREEN       =byte(0x46);
const uint8_t KEY_SCROLL_LOCK        =byte(0x47);
const uint8_t KEY_PAUSE              =byte(0x48);
const uint8_t KEY_INSERT             =byte(0x49);
const uint8_t KEY_HOME               =byte(0x4A);
const uint8_t KEY_PAGE_UP            =byte(0x4B);
const uint8_t KEY_DELETE             =byte(0x4C);
const uint8_t KEY_END                =byte(0x4D);
const uint8_t KEY_PAGE_DOWN          =byte(0x4E);
const uint8_t KEY_ARROW_RIGHT        =byte(0x4F);
const uint8_t KEY_ARROW_LEFT         =byte(0x50);
const uint8_t KEY_ARROW_DOWN         =byte(0x51);
const uint8_t KEY_ARROW_UP           =byte(0x52);
const uint8_t KEY_NUM_LOCK           =byte(0x53);
const uint8_t KEY_KEYPAD_DIVIDE      =byte(0x54);
const uint8_t KEY_KEYPAD_MULTIPLY    =byte(0x55);
const uint8_t KEY_KEYPAD_SUBTRACT    =byte(0x56);
const uint8_t KEY_KEYPAD_ADD         =byte(0x57);
const uint8_t KEY_KEYPAD_ENTER       =byte(0x58);
const uint8_t KEY_KEYPAD_1           =byte(0x59);
const uint8_t KEY_KEYPAD_2           =byte(0x5A);
const uint8_t KEY_KEYPAD_3           =byte(0x5B);
const uint8_t KEY_KEYPAD_4           =byte(0x5C);
const uint8_t KEY_KEYPAD_5           =byte(0x5D);
const uint8_t KEY_KEYPAD_6           =byte(0x5E);
const uint8_t KEY_KEYPAD_7           =byte(0x5F);
const uint8_t KEY_KEYPAD_8           =byte(0x60);
const uint8_t KEY_KEYPAD_9           =byte(0x61);
const uint8_t KEY_KEYPAD_0           =byte(0x62);
const uint8_t KEY_KEYPAD_DECIMAL     =byte(0x63);
const uint8_t KEY_EUROPE_2           =byte(0x64);
const uint8_t KEY_APPLICATION        =byte(0x65);
const uint8_t KEY_POWER              =byte(0x66);
const uint8_t KEY_KEYPAD_EQUAL       =byte(0x67);
const uint8_t KEY_F13                =byte(0x68);
const uint8_t KEY_F14                =byte(0x69);
const uint8_t KEY_F15                =byte(0x6A);
const uint8_t KEY_CONTROL_LEFT       =byte(0xE0);
const uint8_t KEY_SHIFT_LEFT         =byte(0xE1);
const uint8_t KEY_ALT_LEFT           =byte(0xE2);
const uint8_t KEY_GUI_LEFT           =byte(0xE3);
const uint8_t KEY_CONTROL_RIGHT      =byte(0xE4);
const uint8_t KEY_SHIFT_RIGHT        =byte(0xE5);
const uint8_t KEY_ALT_RIGHT          =byte(0xE6);
const uint8_t KEY_GUI_RIGHT          =byte(0xE7);

/*Init Variables*********************************************************************************************************/
/* To add more modes of operation, add additional lines to each array below. Mode[] contains the names of each array.
 *  Each entry will be shown on the LCD display. Each line can seperated by the ~. Text before the ~ will be shown on
 *  line 0 and text afterward will be shown on line 1. Also, line 0 text will be shown on line one in mode select.
 */
const String mode[]={
    "Starting Up...",
    "Ready to Pair...",
    "Mode Select:",
    "KODI~Video Controls",
    "MAME Standard 1~Player 1",
    "MAME Standard 2~Player 2",
    "MAME Alternate 1~Player 1",
    "MAME Alternate 2~Player 2"
};

/* Buttons to Keycodes for each mode - if button is not used to send keycode, set it to hex zero. The keycodes
 *  to text notes can be found on the Keycodes tab. The array is aligned with the folowing buttons:
 *  UP, DOWN, LEFT, RIGHT, ACTION1, ACTION2, ACTION3, ACTION4, ACTION5, ACTION6, PLAYER1, PLAYER2, WHITE, ESCAPE
 */
const uint8_t keyModes[][14] = {
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0}, //Starup Mode
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0}, //Pairing Mode
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0}, //Mode Select Keys
    {KEY_ARROW_UP, KEY_ARROW_DOWN, KEY_ARROW_LEFT, KEY_ARROW_RIGHT, KEY_B, KEY_P, KEY_F, KEY_P, KEY_S, KEY_NONE, KEY_RETURN, KEY_C, KEY_BACKSPACE, KEY_ESCAPE}, //Kodi 
    {KEY_ARROW_UP, KEY_ARROW_DOWN, KEY_ARROW_LEFT, KEY_ARROW_RIGHT, KEY_CONTROL_LEFT, KEY_ALT_LEFT, KEY_SPACE, KEY_SHIFT_LEFT, KEY_Z, KEY_X, KEY_1, KEY_2, KEY_5, KEY_ESCAPE}, //MAME Std P1
    {KEY_R, KEY_F, KEY_D, KEY_G, KEY_A, KEY_S, KEY_Q, KEY_W, KEY_NONE, KEY_NONE, KEY_1, KEY_2, KEY_6, KEY_ESCAPE}, //MAME Std P2
    {KEY_ARROW_DOWN, KEY_ARROW_UP, KEY_ARROW_RIGHT, KEY_ARROW_LEFT, KEY_SHIFT_LEFT, KEY_Z, KEY_X, KEY_CONTROL_LEFT, KEY_ALT_LEFT, KEY_SPACE, KEY_1, KEY_2, KEY_5, KEY_ESCAPE}, //MAME Alt P1
    {KEY_F, KEY_R, KEY_G, KEY_D, KEY_W, KEY_NONE, KEY_NONE, KEY_A, KEY_S, KEY_Q, KEY_1, KEY_2, KEY_6, KEY_ESCAPE}  //MAME Alt P2
};

const uint8_t keyModifiers[][14] = {
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0}, //Starup Mode
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0}, //Pairing Mode
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0}, //Mode Select Keys
    {0,0,0,0, MODIFIER_SHIFT_LEFT | MODIFIER_CONTROL_LEFT, MODIFIER_SHIFT_LEFT | MODIFIER_CONTROL_LEFT, MODIFIER_SHIFT_LEFT | MODIFIER_CONTROL_LEFT, MODIFIER_CONTROL_LEFT, MODIFIER_SHIFT_LEFT | MODIFIER_CONTROL_LEFT,0,0,0,0,0}, //Kodi 
    {0,0,0,0, MODIFIER_CONTROL_LEFT, MODIFIER_ALT_LEFT,0, MODIFIER_SHIFT_LEFT,0,0,0,0,0,0}, //MAME Std P1
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0}, //MAME Std P2
    {0,0,0,0, MODIFIER_SHIFT_LEFT,0,0, MODIFIER_CONTROL_LEFT, MODIFIER_ALT_LEFT,0,0,0,0,0}, //MAME Alt P1
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0}  //MAME Alt P2
};

/* Quick mode select (mode + player button) definitions - if the player button does not change modes, set to zero 
 *  otherwise set to the mode index above that the mode will switch to.
 */
const int quickMode[][3] = {
    {0,0,0},  // Center buttons not used for quick mode change
    {0,0,0},  // Center buttons not used for quick mode change
    {0,0,3},  // Center buttons not used for quick mode change
    {0,0,4},  // White button + mode button switches to mode 'MAME Standard~Player 1'
    {0,5,3},  // Player 2 button + mode button switches to mode 'MAME Standard~Player 2' & White button switches to 'Kodi' mode
    {4,0,3},  // Player 1 button + mode button switches to mode 'MAME Standard~Player 1' & White button switches to 'Kodi' mode
    {0,7,3},  // Player 2 button + mode button switches to mode 'MAME Alternate~Player 2' & White button switches to 'Kodi' mode
    {6,0,3}   // Player 1 button + mode button switches to mode 'MAME Alternate~Player 1' & White button switches to 'Kodi' mode
};

int currentMode = 0;              // the current mode of operation
long debounceDelay = 50;          // the debounce time; increase if the output flickers
long backlightDelay = 0;          // the period of time the backlight is left on after exiting mode select - zero for always on
long lowBattFloor = 650;          // Low battery ADC level. (Measured range 870-525)

/*Work Variables*********************************************************************************************************/
byte btnStates[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
byte prevBtnStates[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

int modeCount;                    // Count of total modes
int currentModeSelected = 0;      // In mode select, the current mode selected on screen
int previousMode = 0;             // the previous mode of operation

bool statesChanged=false;         // Have any of the buttons changed state from the previous reading
bool activeButtons=false;         // Are any of the buttons currently pressed

long lastDebounceTime = 0;        // the last time the output pin was toggled

long lastBacklightTime = 0;       // the countdown timer variable measuring how long the light has been on since save/cancel
bool backlight=HIGH;              // the current state of the LCD backlight

long pairSignalInterval = 0;      // the period of time the pairing signal from the BT module was high
long lastSignalTime = 0;          // the last time the output pin was toggled
int currentPairState = 0;
int lastPairState = 4;            // in the loop function, the last state reported by checkPairing()
bool pairPrevStarted = false;     // has the pairing button already been pressed in mode select

bool powerPrevLow = false;        // latching variable set once when batter drops below lowBattFloor.

/*Object Init***********************************************************************************************************/
// Connect to the LCD via i2c, default address #0 (A0-A2 not jumpered)
LiquidTWI lcd(0);

// Connect to the buttons (mcp23017) via i2c, using address 1
Adafruit_MCP23017 mcp;

// Connect to the Adafruit Bluefruit EZ-Key through a softwareSerial object
SoftwareSerial bfez = SoftwareSerial(10, 11);


/*End Global Init******************************************************************************************************/

// Function to send keycodes and modifiers to the Bluetooth device
void sendKeys(uint8_t modifiers, uint8_t keycode1, uint8_t keycode2 = 0, uint8_t keycode3 = 0, 
                uint8_t keycode4 = 0, uint8_t keycode5 = 0, uint8_t keycode6 = 0) {
  
  bfez.write(0xFD);         // our command
  bfez.write(modifiers);    // modifier - none needed for this application
  bfez.write((byte)0x00);   // 0x00  
  bfez.write(keycode1);     // key code #1
  bfez.write(keycode2);     // key code #2
  bfez.write(keycode3);     // key code #3
  bfez.write(keycode4);     // key code #4
  bfez.write(keycode5);     // key code #5
  bfez.write(keycode6);     // key code #6
}

// Function triggered with pairing signal from bluetooth device activates pin iPairSignal
void handlePairSignal(){
  static long timeSinceOn;  // variable used to mark the list rising change of the pin iPairSignal

  // If the pairing signal just went high from low then capture that time it went high for later use
  if (digitalRead(iPairSignal) == HIGH){
    timeSinceOn = millis();
    // update the pairing led to match the signal from the bluetooth device
    digitalWrite(oPairLED, HIGH);
  }
  else{
    // if the pairing signal just went low from high, calculate the length of time it was high and set global variable
    pairSignalInterval = (millis()-timeSinceOn);
    // update the pairing led to match the signal from the bluetooth device
    digitalWrite(oPairLED, LOW);
  }
}

// Function used to decode the pairing signal from the bluetooth device.
int checkPairing() {
  // Determine the current pairing state based on how long the pairSignalInterval as calculated in handlePairSignal()
  Serial.println(pairSignalInterval);
  
  if ((pairSignalInterval < 120) || (pairSignalInterval > 14000)){
    // Trying to connect
    return 2;
  }

  if ((pairSignalInterval > 150) & (pairSignalInterval < 600)){
    // Pairing mode
    return 0;
  }

  if ((pairSignalInterval > 1500) & (pairSignalInterval < 2500)){
    // Paired
    return 3;
  }

  // if none of the above applies then return "unknown" status of 1
  return 1;
}

// Function used to read the current button states through the mcp object
void getButtons(){
  uint16_t pinstates = 0;
  activeButtons = false;
  statesChanged = false;
  
  pinstates=mcp.readGPIOAB();             // Read all of the I/O pins on the MCP at once
  for(int i=0;i<15;i++){                  // Parce each bit of the returned result into an array
    btnStates[i]=bitRead(pinstates,i);
    if(btnStates[i] == 1)                 // check to see if any of the buttons are active 
      activeButtons = true;
    if(prevBtnStates[i] != btnStates[i])  // check to see if any of the current states have actually change since last reading
      statesChanged = true;
  }
}

// Function to display the current operational mode on the LCD
void displayMode(){

  // Clear the screen and prep work variables
  lcd.clear();
  String lineOne = "";
  String lineTwo = "";

  // Check the text of the current mode to see if there is a first line / second line devider (~)
  int c=mode[currentMode].indexOf("~");
  int l=mode[currentMode].length();

  switch(c){
    case -1:  // No divider - everything should go on the top line
      lineOne=mode[currentMode];
      break;
    case 0:   // The devider is at the beginning of the text - Everything goes on the second line
      for(int h=1;h<l;h++)
        lineTwo+=mode[currentMode].charAt(h);
      break;
    default:  // The divider is somewhere in the middle so divde the text accordingly
      for(int j=0;j<c;j++)
        lineOne+=mode[currentMode].charAt(j);
      
      for(int k=(c+1);k<l;k++)
        lineTwo+=mode[currentMode].charAt(k);
      break;    
  }

  // Set the backlight and working variable to high
  lcd.setBacklight(HIGH);
  backlight=HIGH;

  // Write the text to the display
  lcd.setCursor(0,0);
  lcd.print(lineOne);
  lcd.setCursor(0,1);
  lcd.print(lineTwo);

  // Set global time keeping variable for when the backlight was turned on for later use
  lastBacklightTime = millis();
}

// Function to display the current selected mode on line 2 of the LCD in mode select mode
void displayMSMode(int modeIndex){

  // Clear the screen and prep work variables
  lcd.clear();
  String lineOne = mode[2];
  String lineTwo = "";

  // Check the text of the current mode to see if there is a first line / second line devider (~)
  int c=mode[modeIndex].indexOf("~");
  int l=mode[modeIndex].length();

  switch(c){
    case -1:  // No divider - everything should go on the bottom line
      lineTwo=mode[modeIndex];
      break;
    case 0:   // The devider is at the beginning of the text - Everything goes on the second line
      for(int h=1;h<l;h++)
        lineTwo+=mode[modeIndex].charAt(h);
      break;
    default:  // The divider is somewhere in the middle so divde the text and only write the first line of text to line two
      for(int j=0;j<c;j++)
        lineTwo+=mode[modeIndex].charAt(j);
      break;    
  }

  // Set the backlight and working variable to high
  lcd.setBacklight(HIGH);
  backlight=HIGH;

  // Write the text to the display
  lcd.setCursor(0,0);
  lcd.print(lineOne);
  lcd.setCursor(0,1);
  lcd.print(lineTwo);

  // Set global time keeping variable for when the backlight was turned on for later use
  lastBacklightTime = millis();
}


/*Primary Functions************************************************************************************************/


void setup() {

  // Initialize serial connection with the EZ-Key
  bfez.begin(9600);

  // Set pin types
  pinMode(iPairSignal, INPUT);
  pinMode(oPairBtn, OUTPUT);
  pinMode(oPairLED, OUTPUT);
  pinMode(oLowBatLED, OUTPUT);
  pinMode(iVBat, INPUT);
  pinMode(oPowerBtnLED, OUTPUT);
  pinMode(oModeBtnLED, OUTPUT);
  pinMode(oEscapeBtnLED, OUTPUT);

  // Set the pairing pin low to keep the bluetooth from entering pairing mode accidentally
  digitalWrite(oPairBtn, LOW);

  // Set the interrupt for the incoming pairing signal
  attachInterrupt(digitalPinToInterrupt(iPairSignal), handlePairSignal, CHANGE);

  // Get the total number of defined modes above
  modeCount = (((sizeof(keyModes) / sizeof(uint8_t))/14)-1);
  
  // set up the LCD's number of rows and columns: 
  lcd.begin(16, 2);

  // Initiate communications with MCP on address 1
  mcp.begin(1);

  // Set the pin types for the first 4 inputs (pins 0-3) as active high inputs (internal pullup set low)
  for(int h=0;h<4;h++){
    mcp.pinMode(h, INPUT);
    mcp.pullUp(h, LOW);
  }

  // Set the pin types for the next 9 inputs (pins 4-12) as active low inputs
  for(int i=4;i<13;i++){
    mcp.pinMode(i, INPUT);
    mcp.pullUp(i, HIGH);  // turn on a 100K pullup internally
  }

  // Set pin 13 and 14 as active high
  mcp.pinMode(13, INPUT);
  mcp.pullUp(13, LOW);
  mcp.pinMode(14, INPUT);
  mcp.pullUp(14, LOW);

  // Check LEDs
  digitalWrite(oEscapeBtnLED, HIGH);
  digitalWrite(oModeBtnLED, HIGH);
  digitalWrite(oLowBatLED, HIGH);
  digitalWrite(oPairLED, HIGH);
  // digitalWrite(oPowerBtnLED, HIGH);
  delay(750);
  // Turn LEDs except power LED back off
  digitalWrite(oEscapeBtnLED, LOW);
  digitalWrite(oModeBtnLED, LOW);
  digitalWrite(oLowBatLED, LOW);
  digitalWrite(oPairLED, LOW);

  // Check to see if the escape button is pressed on startup which bypasses pairing tests
  getButtons();
  // if the states have changed
  if (statesChanged){
    // and at least one of the buttons is active...
    if (activeButtons){
      // Escape Key pressed
      if (btnStates[13] == 1){
        currentMode = 2;
        currentModeSelected = 3;
        displayMode();
        digitalWrite(oModeBtnLED, HIGH);
        displayMSMode(currentModeSelected);
      }
      else{
        displayMode();
      }
    }
    else{
      displayMode();
    }
  }
  else{
    displayMode();
  }

}

void loop() {
  
  // if we're not in Mode Select and the backlight is still on and it has been longer than the 
  // value of backlightDelay, then turn the LCD backlight off - if delay is set to 0 then leave it on
  if (backlightDelay > 0){
    if((currentMode>2) && (backlight == HIGH) && ((millis() - lastBacklightTime)>backlightDelay)){
      lcd.setBacklight(LOW);
      backlight=LOW;
      lastBacklightTime = millis();
    }
  }

  // Check to see if the battery is low. if it is, then set a latch variable so that the low batt LED
  // doesn't fluctuate as the reading fluctuates. 
  if (powerPrevLow == false){
    if (analogRead(iVBat) < lowBattFloor){
      digitalWrite(oLowBatLED, HIGH);
      powerPrevLow = true;
    }
  }

  // The branches of this switch tree represent the operational modes - the first 3 are the built in functions 
  // followed by the the user defined additional modes. 
  switch (currentMode){
    case 0:   // Start up mode
      // Check Bluetooth Pairing
      currentPairState = checkPairing();
      if (currentPairState != lastPairState){
        switch (currentPairState){
          case 0:   // Pairing
            currentMode = 1;
            displayMode();
            digitalWrite(oEscapeBtnLED,HIGH);
            lastPairState = currentPairState;
            break;
            
          case 1:   // Unknown Status
            lcd.setCursor(0,1);
            lcd.print("                ");
            lcd.setCursor(0,1);
            lcd.print("Bluetooth Error");
            digitalWrite(oEscapeBtnLED,HIGH);
            lastPairState = currentPairState;
            break;
            
          case 2:   // Paired but attempting to connect
            lcd.setCursor(0,1);
            lcd.print("                ");
            lcd.setCursor(0,1);
            lcd.print("BT Connecting...");
            digitalWrite(oEscapeBtnLED,HIGH);
            lastPairState = currentPairState;
            break;
            
          case 3:   // Paired - move on to select operational mode
            currentMode = 2;
            currentModeSelected = 3;
            displayMode();
            digitalWrite(oModeBtnLED, HIGH);
            digitalWrite(oEscapeBtnLED,LOW);
            displayMSMode(currentModeSelected);
            lastPairState = currentPairState;
            break;
          
          default:
            lcd.setCursor(0,1);
            lcd.print("                ");
            lcd.setCursor(0,1);
            lcd.print("Bluetooth Error");
            digitalWrite(oEscapeBtnLED,HIGH);
            lastPairState = currentPairState;
            break;
        }
      }
      break;
    case 1:   // Pairing Mode
      // Check Bluetooth Pairing
      currentPairState = checkPairing();
      if (currentPairState != lastPairState){
        switch (currentPairState){
          case 0:   // Pairing
            lastPairState = currentPairState;
            break;
            
          case 1:   // Unknown Status
            lcd.setCursor(0,1);
            lcd.print("                ");
            lcd.setCursor(0,1);
            lcd.print("Bluetooth Error");
            digitalWrite(oEscapeBtnLED,HIGH);
            lastPairState = currentPairState;
            break;
            
          case 2:   // Paired but attempting to connect
            lcd.setCursor(0,1);
            lcd.print("                ");
            lcd.setCursor(0,1);
            lcd.print("BT Connecting...");
            digitalWrite(oEscapeBtnLED,HIGH);
            lastPairState = currentPairState;
            break;
            
          case 3:   // Paired - move on to select operational mode
            currentMode = 2;
            currentModeSelected = 3;
            displayMode();
            digitalWrite(oModeBtnLED, HIGH);
            digitalWrite(oEscapeBtnLED,LOW);
            displayMSMode(currentModeSelected);
            lastPairState = currentPairState;
            break;
          
          default:
            lcd.setCursor(0,1);
            lcd.print("                ");
            lcd.setCursor(0,1);
            lcd.print("Bluetooth Error");
            digitalWrite(oEscapeBtnLED,HIGH);
            lastPairState = currentPairState;
            break;
        }
      }
      break;
      
    case 2:   // Mode Select
      // wait for the debounce time to expire
      if ((millis() - lastDebounceTime) > debounceDelay) {
        // Get the readings from the buttons which sets the statesChanged and activeButtons global variables
        getButtons();
        // if the states have changed since the last reading...
        if (statesChanged){
          // and at least one of the buttons is active...
          if (activeButtons){
            // Handle the current keystrokes as only 6 are allowed in this mode
            // Escape Key pressed
            // If there was a previously selected user mode then switch back to that previous mode, otherwise ignore
            if ((btnStates[13] == 1) && (previousMode > 2)){
                currentMode=previousMode;
                displayMode();
                digitalWrite(oModeBtnLED, LOW);
                // make current the previous states and time
                lastDebounceTime = millis();
                for(int i=0;i<15;i++){
                  prevBtnStates[i]=btnStates[i];
                }
                break;
            }
            
            // Player 1 key pressed 
            // Switch to the Quick mode defined for the last operational mode
            if ((btnStates[10] == 1) && (previousMode > 2)){
              if (quickMode[previousMode][0] != 0){
                currentMode=quickMode[previousMode][0];
                displayMode();
                digitalWrite(oModeBtnLED, LOW);
                // make current the previous states and time
                lastDebounceTime = millis();
                for(int i=0;i<15;i++){
                  prevBtnStates[i]=btnStates[i];
                }
                break;
              }
            }
            
            // Player 2 key pressed
            // Switch to the Quick mode defined for the last operational mode
            if ((btnStates[11] == 1) && (previousMode > 2)){
              if (quickMode[previousMode][1] != 0){
                currentMode=quickMode[previousMode][1];
                displayMode();
                digitalWrite(oModeBtnLED, LOW);
                // make current the previous states and time
                lastDebounceTime = millis();
                for(int i=0;i<15;i++){
                  prevBtnStates[i]=btnStates[i];
                }
                break;
              }
            }

            // White key pressed
            // Switch to the Quick mode defined for the last operational mode
            if ((btnStates[12] == 1) && (previousMode > 2)){
              if (quickMode[previousMode][2] != 0){
                currentMode=quickMode[previousMode][2];
                displayMode();
                digitalWrite(oModeBtnLED, LOW);
                // make current the previous states and time
                lastDebounceTime = millis();
                for(int i=0;i<15;i++){
                  prevBtnStates[i]=btnStates[i];
                }
                break;
              }
            }
            
            // UP key pressed
            // Display the previous user defined mode 
            if (btnStates[0] == 1){
              if ((currentModeSelected -1) >= 3){
                currentModeSelected--;
                displayMSMode(currentModeSelected);
              }
              else{
                currentModeSelected = modeCount;
                displayMSMode(currentModeSelected);
              }
            }
            
            // DOWN key pressed
            // Display the next user defined mode
            if (btnStates[1] == 1){
              if ((currentModeSelected +1) <= modeCount){
                currentModeSelected++;
                displayMSMode(currentModeSelected);
              }
              else{
                currentModeSelected = 3;
                displayMSMode(currentModeSelected);
              }
            }

            // Action1 button pressed (pair)
            // Active the pairing pin to the bluetooth device, switch to pairing system mode
            // and then disable once it registers Pairing command
            if (btnStates[4] == 1 || pairPrevStarted == true){
              if (pairPrevStarted == false){
                digitalWrite(oPairBtn,HIGH);
                pairPrevStarted = true;
              }
              else{
                if (checkPairing() == 1){
                  digitalWrite(oPairBtn,LOW);
                  digitalWrite(oEscapeBtnLED,HIGH);
                  currentMode = 1;
                  displayMode();
                  pairPrevStarted = false;
                  break;
                }
              }
            }
            
            // Mode (save) key pressed
            // Change to the currently displayed (selected) user defined mode
            if (btnStates[14]==1){
              currentMode=currentModeSelected;
              displayMode();
              digitalWrite(oModeBtnLED, LOW);
              // make current the previous states and time
              lastDebounceTime = millis();
              for(int i=0;i<15;i++){
                prevBtnStates[i]=btnStates[i];
              }
              break;
            }
            
          }
          // make current the previous states and time
          lastDebounceTime = millis();
          for(int i=0;i<15;i++){
            prevBtnStates[i]=btnStates[i];
          }
        }
      } 
      break;
    default:  // All run modes
      // wait for the debounce time to expire
      if ((millis() - lastDebounceTime) > debounceDelay) {
        // Get the readings from the buttons which sets the statesChanged and activeButtons global variables
        getButtons();
        // if the states have changed since the last reading...
        if (statesChanged){
          // and at least one of the buttons is active...
          if (activeButtons){
            // send the proper keystrokes for the current mode
            uint8_t keycodes[6] = {0,0,0,0,0,0};  // array to keep up with the first 6 active buttons to send
            uint8_t modifiers = 0;   // variable to store all of the key modfiers for the keys used. the moifiers are ORed together
            int c =0;   // counter for how many keycodes we've used so far
            if (btnStates[14]==0){
              for(int i=0;i<14;i++){
                // if the current button is active and we have not already identified 6 simultaneously pressed buttons
                // then add the current mode's button key to the key list
                if ((btnStates[i]) && (c<6)){
                  keycodes[c] = keyModes[currentMode][i];
                  modifiers = (modifiers | keyModifiers[currentMode][i]);
                  c++;
                }
              }
              sendKeys(modifiers,keycodes[0],keycodes[1],keycodes[2],keycodes[3],keycodes[4],keycodes[5]);
            }
            else{
              // Switching to Mode select...
              // Make sure to tell the computer that no keys are currently pressed
              sendKeys(0,0);
              // Set the previous mode to the current mode in case Mode select is cancelled
              previousMode=currentMode;
              currentModeSelected = currentMode;
              // Change the current mode to mode select
              currentMode=2;
              // display the new mode
              displayMode();
              digitalWrite(oModeBtnLED, HIGH);
              displayMSMode(currentModeSelected);
            }
          }
          else{
            // if the buttons are all inactive, then send "no button" 
            sendKeys(0,0);
          }
          // make current the previous states and time
          lastDebounceTime = millis();
          for(int i=0;i<15;i++){
            prevBtnStates[i]=btnStates[i];
          }
        }
      } 
      break;
  }
}

